/*********************************************************************
*                SEGGER Microcontroller GmbH & Co. KG                *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2016 SEGGER Microcontroller GmbH & Co. KG         *
*                                                                    *
*     Internet: segger.com   Support: support_embos@segger.com       *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.16                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_AT91SAM7S.c
          for ATMEL AT91SAM7S / AT91SAM7X CPUs

Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#ifdef __ICCARM__
  #include <intrinsics.h>
#endif

#ifdef __GNUC__
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable"  // Avoid warning about unused variables
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

#define ALLOW_NESTED_INTERRUPTS 1   // Caution: Nesting interrupts will cause higher stack load on system stack CSTACK

/*********************************************************************
*
*       Clock frequency settings
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS 47923200uL        /* Assume 47.92MHz              */
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS)   /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART (OS_FSYS)    /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000)
#endif

#define OS_TIMER_PRESCALE (16) /* prescaler for system timer fixed to 16 */

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_ENABLE            // Global enable of communication
  #define OS_VIEW_ENABLE    (1)     // Default: on
#endif

#ifndef   OS_VIEW_USE_UART          // If set, UART will be used
  #define OS_VIEW_USE_UART  (0)     // Default: 0 => DCC is used
#endif                              // if OS_VIEW_ENABLE is on

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently UART0 and UART1 are supported and the standard
*       setup enables UART 0 per default
*       When using DCC for communication, the UART is not used for embOSView,
*       regardless the OS_UART settings
*/
#ifndef   OS_UART
  #define OS_UART (0)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (38400)
#endif

/****** End of configurable options *********************************/

#define OS_USE_DCC     (OS_VIEW_ENABLE && (OS_VIEW_USE_UART == 0))

#define OS_UART_USED   (OS_VIEW_ENABLE && (OS_VIEW_USE_UART != 0)) && ((OS_UART == 0) || (OS_UART == 1))

#if  (OS_UART_USED || OS_USE_DCC)
  #define OS_COM_INIT() OS_COM_Init()
#else
  #define OS_COM_INIT()
#endif

#define OS_TIMER_RELOAD ((OS_PCLK_TIMER/OS_TIMER_PRESCALE/OS_TICK_FREQ) - 1)
#if (OS_TIMER_RELOAD >= 0x00100000)
  #error "PIT timer can not be used, please check configuration"
#endif

/*********************************************************************
*
*       Check configuration
*
**********************************************************************
*/

#ifndef   DEBUG     /* Should normally be defined as project option */
  #define DEBUG  (0)
#endif

#if ((defined(__ICCARM__) && DEBUG) || OS_USE_DCC)
  #include "JLINKDCC.h"
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/

/*      USART, used for OSView communication */
#define USART0_BASE_ADDR  (0xFFFC0000)
#define USART1_BASE_ADDR  (0xFFFC4000)

/*      Debug unit */
#define DBGU_BASE_ADDR    (0xFFFFF200)
#define DBGU_IMR          (*(volatile OS_U32*) (DBGU_BASE_ADDR + 0x10)) /* Interrupt Mask Register */
#define DBGU_SR           (*(volatile OS_U32*) (DBGU_BASE_ADDR + 0x14)) /* Channel Status Register */
#define DBGU_CPUID        (*(volatile OS_U32*) (DBGU_BASE_ADDR + 0x40)) /* Chip ID Register        */
#define DBGU_COMMRX       ((OS_U32)1 << 31)
#define DBGU_COMMTX       (1uL << 30)
#define DBGU_RXBUFF       (1uL << 12)
#define DBGU_TXBUFE       (1uL << 11)
#define DBGU_TXEMPTY      (1uL <<  9)
#define DBGU_PARE         (1uL <<  7)
#define DBGU_FRAME        (1uL <<  6)
#define DBGU_OVRE         (1uL <<  5)
#define DBGU_ENDTX        (1uL <<  4)
#define DBGU_ENDRX        (1uL <<  3)
#define DBGU_TXRDY        (1uL <<  1)
#define DBGU_RXRDY        (1uL <<  0)
#define DBGU_MASK_ALL     (DBGU_COMMRX | DBGU_COMMTX  | DBGU_RXBUFF |  \
                           DBGU_TXBUFE | DBGU_TXEMPTY | DBGU_PARE   |  \
                           DBGU_FRAME  | DBGU_OVRE    | DBGU_ENDTX  |  \
                           DBGU_ENDRX  | DBGU_TXRDY   | DBGU_RXRDY)

/*      Reset controller */
#define RSTC_BASE_ADDR    (0xFFFFFD00)
#define RSTC_CR       (*(volatile OS_U32*) (RSTC_BASE_ADDR + 0x00))
#define RSTC_SR       (*(volatile OS_U32*) (RSTC_BASE_ADDR + 0x04))
#define RSTC_MR       (*(volatile OS_U32*) (RSTC_BASE_ADDR + 0x08))
#define RSTC_BODIEN   (1uL << 16)  /* Brownout interrupt enable   */
#define RSTC_URSTIEN  (1uL <<  4)  /* User reset interrupt enable */
#define RSTC_BODSTS   (1uL <<  1)  /* Brownout status             */
#define RSTC_URSTS    (1uL <<  0)  /* User reset status           */

/*      Real time timer */
#define RTT_BASE_ADDR     (0xFFFFFD20)
#define RTT_MR        (*(volatile OS_U32*) (RTT_BASE_ADDR + 0x00))
#define RTT_SR        (*(volatile OS_U32*) (RTT_BASE_ADDR + 0x0C))
#define RTT_RTTINCIEN (1uL << 17)
#define RTT_ALMIEN    (1uL << 16)
#define RTT_RTTINC    (1uL << 1)
#define RTT_ALMS      (1uL << 0)

/*      Periodic interval timer */
#define PIT_BASE_ADDR     (0xFFFFFD30)
#define PIT_MR        (*(volatile OS_U32*) (PIT_BASE_ADDR + 0x00))
#define PIT_SR        (*(volatile OS_U32*) (PIT_BASE_ADDR + 0x04))
#define PIT_PIVR      (*(volatile OS_U32*) (PIT_BASE_ADDR + 0x08))
#define PIT_PIIR      (*(volatile OS_U32*) (PIT_BASE_ADDR + 0x0C))

/*      Watchdog */
#define WDT_BASE_ADDR     (0xFFFFFD40)
#define WDT_CR        (*(volatile OS_U32*) (WDT_BASE_ADDR + 0x00))
#define WDT_MR        (*(volatile OS_U32*) (WDT_BASE_ADDR + 0x04))
#define WDT_SR        (*(volatile OS_U32*) (WDT_BASE_ADDR + 0x08))
#define WDT_WDFIEN    (1uL << 12) /* Watchdog interrupt enable flag in mode register */
#define WDT_WDERR     (1uL <<  1) /* Watchdog error status flag                      */
#define WDT_WDUNF     (1uL <<  0) /* Watchdog underflow status flag                  */

/*      Memory controller */
#define MC_BASE_ADDR      (0xFFFFFF00)
#define MC_FMR0       (*(volatile OS_U32*) (MC_BASE_ADDR + 0x60))
#define MC_FMR1       (*(volatile OS_U32*) (MC_BASE_ADDR + 0x70))

#define MC_FWS_0FWS   ((unsigned int) 0x0 <<  8) /* (MC) 1 cycle for Read,  2 for Write operations */
#define MC_FWS_1FWS   ((unsigned int) 0x1 <<  8) /* (MC) 2 cycles for Read, 3 for Write operations */
#define MC_FWS_2FWS   ((unsigned int) 0x2 <<  8) /* (MC) 3 cycles for Read, 4 for Write operations */
#define MC_FWS_3FWS   ((unsigned int) 0x3 <<  8) /* (MC) 4 cycles for Read, 4 for Write operations */
#if   (OS_FSYS <= 30000000)
  #define MC_FWS (MC_FWS_0FWS)
#elif (OS_FSYS <= 50000000)
  #define MC_FWS (MC_FWS_1FWS)
#else
  #error "Please examine settings for memory controller, Flash wait states"
#endif

/*      PIO control register */
#define PIOA_BASE_ADDR    (0xfffff400)
#define PIOA_PDR      (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x04)) /* PIOA disable register               */
#define PIOA_ASR      (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x70)) /* PIOA "A" peripheral select register */
#define PIOA_BSR      (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x74)) /* PIOA "B" peripheral select register */

/*      Power management controller */
#define PMC_BASE_ADDR     (0xFFFFFC00)
#define PMC_SCDR      (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x04))  /* System Clock Disable Register */
#define PMC_PCER      (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x10))  /* Peripheral clock enable register */
#define PMC_MOR       (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x20))  /* main oscillator register */
#define PMC_PLLR      (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x2c))  /* PLL register */
#define PMC_MCKR      (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x30))  /* Master clock register */
#define PMC_SR        (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x68))  /* status register */
#define PMC_IMR       (*(volatile OS_U32*) (PMC_BASE_ADDR + 0x6C))  /* interrupt mask register */
#define PMC_PCKRDY2   (1uL << 10)
#define PMC_PCKRDY1   (1uL <<  9)
#define PMC_PCKRDY0   (1uL <<  8)
#define PMC_MCKRDY    (1uL <<  3)
#define PMC_LOCK      (1uL <<  2)
#define PMC_MOSCS     (1uL <<  0)
#define PMC_MASK_ALL  (PMC_PCKRDY2 | PMC_PCKRDY1 | PMC_PCKRDY0 | \
                       PMC_MCKRDY  | PMC_LOCK    | PMC_MOSCS)

/*      Advanced interrupt controller (AIC) */
#define AIC_BASE_ADDR      (0xfffff000)
#define AIC_SMR_BASE_ADDR  (AIC_BASE_ADDR + 0x00)
#define AIC_SVR_BASE_ADDR  (AIC_BASE_ADDR + 0x80)
#define AIC_SVR0      (*(volatile OS_U32*) (AIC_SVR_BASE_ADDR + 0x00))
#define AIC_SVR1      (*(volatile OS_U32*) (AIC_SVR_BASE_ADDR + 0x04))
#define AIC_IVR       (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x100))
#define AIC_ISR       (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x108))
#define AIC_IPR       (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x10c))
#define AIC_IDCR      (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x124))
#define AIC_ICCR      (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x128))
#define AIC_IECR      (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x120))
#define AIC_EOICR     (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x130))
#define AIC_SPU       (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x134))
#define AIC_DCR       (*(volatile OS_U32*) (AIC_BASE_ADDR + 0x138))

/*      AIC interrupt sources and peripheral IDs        */
#define SYSTEM_IRQ_ID  (1)   /* System IRQ ID             */
#define US0IRQ_ID      (6)   /* USART Channel 0 interrupt */
#define US1IRQ_ID      (7)   /* USART Channel 1 interrupt */
#define US2IRQ_ID      (8)   /* USART Channel 2 interrupt */
#define TC0IRQ_ID     (12)   /* Timer Channel 0 interrupt */

#ifndef   NUM_INT_SOURCES
  #define NUM_INT_SOURCES   (32)
#endif

#define INT_PRIORITY_MASK (0x07)
#define NUM_INT_PRIORITIES   (8)

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _HandlePmcIrq(), Power management controller interrupt
*       If not used for application, this handler may be removed
*/
static void _HandlePmcIrq(void) {
#if DEBUG
  OS_U32 IrqSource;
  IrqSource  = PMC_IMR;
  IrqSource &= (PMC_SR & PMC_MASK_ALL);
  if (IrqSource) {  /* PMC interrupt pending? */
    while(1);       /* Not implemented        */
  }
#endif
}

/*********************************************************************
*
*       _HandleRttIrq(), Real time timer interrupt handler
*       If not used for application, this handler may be removed
*/
static void _HandleRttIrq(void) {
#if DEBUG
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RTT_MR & (RTT_RTTINCIEN  | RTT_ALMIEN);
  IrqStatus  = RTT_SR & (RTT_RTTINC | RTT_ALMS);
  if ((IrqStatus & RTT_RTTINC) && (IrqEnabled & RTT_RTTINCIEN )) { /* RTT inc. interrupt pending ? */
    while(1);                /* Not implemented */
  }
  if ((IrqStatus & RTT_ALMS) && (IrqEnabled & RTT_ALMIEN )) {      /* Alarm interrupt pending ? */
    while(1);                /* Not implemented */
  }
#endif
}

/*********************************************************************
*
*       _HandleDbguIrq(), Debug unit interrupt handler
*       If not used for application, this handler may be removed
*/
static void _HandleDbguIrq(void) {
#if DEBUG
  OS_U32 IrqSource;

  IrqSource  = DBGU_IMR;
  IrqSource &= (DBGU_SR & DBGU_MASK_ALL);
  if (IrqSource) { /* Any interrupt pending ? */
    while(1);      /* Not implemented         */
  }
#endif
}

/*********************************************************************
*
*       _HandleRstcIrq(), Reset controller interrupt handler
*       If not used for application, this handler may be removed
*/
static void _HandleRstcIrq(void) {
#if DEBUG
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RSTC_MR & (RSTC_BODIEN | RSTC_URSTIEN);
  IrqStatus  = RSTC_SR & (RSTC_BODSTS | RSTC_URSTS);
  if ((IrqStatus & RSTC_BODSTS) && (IrqEnabled & RSTC_BODIEN )) {  /* Brownout interrupt pending ?   */
    while(1);                /* Not implemented */
  }
  if ((IrqStatus & RSTC_URSTS) && (IrqEnabled & RSTC_URSTIEN )) {  /* User reset interrupt pending ? */
    while(1);                /* Not implemented */
  }
#endif
}

/*********************************************************************
*
*       _HandleWdtIrq(), watchdog timer interrupt handler
*       If not used for application, this handler may be removed
*/
static void _HandleWdtIrq(void) {
#if DEBUG
  OS_U32 IrqStatus;

  IrqStatus = WDT_SR & (WDT_WDERR | WDT_WDUNF);
  if (IrqStatus && (WDT_MR & WDT_WDFIEN)) { /* Watchdog error interrupt pending ? */
    while(1);                               /* Not implemented */
  }
#endif
}

/*********************************************************************
*
*       _DefaultFiqHandler(), a dummy FIQ handler
*
*       This handler is initially written into the AIC FIQ vector.
*       It is called, if the FIQ vector is not initialized,
*       but any interrupting source was assigned to FIQ.
*       May be used during debugging to detect an uninstalled FIQ interrupt
*/
static void _DefaultFiqHandler(void) {
#if DEBUG
  while(1);
#endif
}

/*********************************************************************
*
*       _DefaultIrqHandler, a dummy IRQ handler
*
*       This handler is initially written into all AIC interrupt vectors.
*       It is called, if no interrupt vector was installed for
*       specific interrupt source.
*       May be used during debugging to detect uninstalled interrupts
*/
static void _DefaultIrqHandler(void) {
#if (DEBUG || OS_DEBUG)
  volatile OS_U32 IrqSource;
  IrqSource = AIC_ISR;        // Detect source of uninstalled interrupt
  #if OS_DEBUG
    OS_Error(OS_ERR_ISR_VECTOR);
    return;
  #else
    while(1) {
    }
  #endif
#endif
}

/*********************************************************************
*
*       _SpuriousIrqHandler(), default spurious IRQ handler
*
*       A spurious interrupt is called under one of the following
*       circumstances:
*       1. An external interrupt source is programmed in level-sensitive mode
*          and an active level occurs for only a short time.
*       2. An internal interrupt source is programmed in level sensitive
*          and the output signal of the corresponding embedded peripheral
*          is activated for a short time. (As in the case for the Watchdog.)
*       3. An interrupt occurs just a few cycles before the software begins
*          to mask it, thus resulting in a pulse on the interrupt source.
*       The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if DEBUG
  static OS_U32 SpuriousIRQCnt;
  SpuriousIRQCnt++;
#endif
}

/*********************************************************************
*
*       _OS_SystemIrqhandler()
*       the OS system interrupt, handles OS timer
*/
static void _OS_SystemIrqhandler(void) {
  volatile OS_U32 Dummy;      /* Must not be optimized away         */
  if (PIT_SR & (1uL << 0)) {  /* Timer interupt pending?            */
    Dummy = PIT_PIVR;         /* Reset interrupt pending condition  */
    OS_TICK_Handle();         /* Call OS tick handler               */
#if ((defined(__ICCARM__) && DEBUG) || OS_USE_DCC)
    JLINKDCC_Process();
#endif
  }
  /* Call to following handlers may be removed if not used by application */
  _HandlePmcIrq();
  _HandleRttIrq();
  _HandleDbguIrq();
  _HandleRstcIrq();
  _HandleWdtIrq();
}

/*********************************************************************
*
*       _InitVIC()
*
*       Initialize interupt controller by setting default vectors
*       and clearing all interrupts
*/
static void _InitVIC(void) {
  int  i;
  OS_ISR_HANDLER** papISR;

  AIC_IDCR = 0xFFFFFFFF;                     /* Disable all interrupts     */
  AIC_ICCR = 0xFFFFFFFF;                     /* Clear all interrupts       */
  AIC_SVR0 = (int) _DefaultFiqHandler;       /* Dummy FIQ handler          */
  AIC_SPU  = (int) _SpuriousIrqHandler ;     /* Dummy spurious handler     */
  papISR   = (OS_ISR_HANDLER**) AIC_SVR_BASE_ADDR;
  for (i = 1; i < NUM_INT_SOURCES; i++)  {   /* Initially set all sources  */
    *(papISR + i) = &_DefaultIrqHandler;     /* to dummy irq handler       */
  }
  for (i = 0; i < NUM_INT_PRIORITIES; i++) {
    AIC_EOICR = 0;                           /* Reset interrupt controller */
  }
#if DEBUG                                    /* For debugging activate AIC protected mode */
  AIC_DCR |= 0x01;                           /* Enable AIC protected mode  */
#endif
}

/*********************************************************************
*
*       _InitRTT()
*
* Function description
*   Initialize Real time timer.
*
* NOTES: (1) Not used by embOS, we disable interrupts here to avoid unhandled interrupts
*            May be modified by user if RTT is required for application.
*/
static void _InitRTT(void) {
  //
  // Disable Real-Time Timer interrupts
  //
  RTT_MR &= ~(RTT_RTTINCIEN  | RTT_ALMIEN);
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*   Returns the current hardware timer count value
*
* Return value
*   Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return (PIT_PIIR & 0x000FFFFF);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*   Returns if the hardware timer interrupt pending flag is set
*
* Return value
*   == 0; Interrupt pending flag not set
*   != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (PIT_PIIR & 0xFFF00000);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 1u, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_IncDI();
  /* Initialize PIT as OS timer, enable timer + timer interrupt */
  PIT_MR = ((OS_TIMER_RELOAD & 0x000FFFFF) | (1uL << 25) | (1uL << 24));
  OS_ARM_InstallISRHandler(SYSTEM_IRQ_ID, _OS_SystemIrqhandler);
  OS_ARM_EnableISR(SYSTEM_IRQ_ID);
  //
  // Setup values for usec precise system time functions
  //
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Initialize the optional UART for OS viewer
  //
  OS_COM_INIT();
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
#if (DEBUG == 0)
    PMC_SCDR = 1;       // Switch off CPU clock to save power
#endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*       This routine is required for task-info via OSView or high
*       resolution time measurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  unsigned int t_cnt;
  OS_U32 time ;

  t_cnt = PIT_PIIR;            /* Read current timer value   */
  time  = OS_GetTime32();      /* Read current OS time    */
  if (t_cnt & 0xFFF00000) {    /* Timer Interrupt pending ?  */
    time  += (t_cnt >> 20);    /* Adjust result              */
    t_cnt &= 0x000FFFFF;
  }
  return (OS_TIMER_RELOAD * time) + t_cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagnostics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {

#define OS_TIMER_FREQ (OS_PCLK_TIMER/OS_TIMER_PRESCALE)

#if (OS_TIMER_FREQ/1000000 * 1000000 == OS_TIMER_FREQ)
  return Cycles/(OS_PCLK_TIMER/OS_TIMER_PRESCALE/1000000);
#else
  long long t;
  t = Cycles;
  t *= 1000000;
  t /= OS_TIMER_FREQ;
  return (OS_U32)t;
#endif
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

#if OS_UART_USED

  #if (OS_UART == 0)
    #define OS_UART_BASE_ADDR   (USART0_BASE_ADDR)
    #define OS_UART_ID          (US0IRQ_ID)
    #define OS_UART_RX_PIN      (5)
    #define OS_UART_TX_PIN      (6)
  #elif (OS_UART == 1)
    #define OS_UART_BASE_ADDR   (USART1_BASE_ADDR)
    #define OS_UART_ID          (US1IRQ_ID)
    #define OS_UART_RX_PIN      (21)
    #define OS_UART_TX_PIN      (22)
  #endif

  #define PIO_PDR_USART ((1uL << OS_UART_RX_PIN) | (1uL << OS_UART_TX_PIN))

  #define US_CR   (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x00))
  #define US_MR   (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x04))
  #define US_IER  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x08))
  #define US_IDR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x0c))
  #define US_IMR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x10))
  #define US_CSR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x14))
  #define US_RHR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x18))
  #define US_THR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x1c))
  #define US_BRGR (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x20))
  #define US_RTOR (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x24))
  #define US_TTGR (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x28))

  #define US_TPR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x108))
  #define US_TCR  (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x10c))
  #define US_PTCR (*(volatile OS_U32*)(OS_UART_BASE_ADDR + 0x120))

  #define US_RXRDY   (1uL << 0)         /* Rx status flag                  */
  #define US_TXRDY   (1uL << 1)         /* Tx RDY Status flag              */
  #define US_TXEMPTY (1uL << 9)         /* Tx EMPTY Status flag            */
  #define US_ENDTX   (1uL << 4)         /* Tx end flag                     */
  #define USART_RX_ERROR_FLAGS (0xE0) /* Parity, framing, overrun error  */

  #define OS_UART_PRIO    (0x00)      /* Lowest priority for UART interrupts */
  #define US_TX_INT_FLAG  (US_ENDTX)

/*********************************************************************
*
*       OS_COM_ISR_Usart() OS USART interrupt handler
*       handles both, Rx and Tx interrupt
*/
static void OS_COM_ISR_Usart(void) {
  volatile OS_U32 Dummy;                   /* Must not be optimized away */
  int UsartStatus;

  UsartStatus = US_CSR;                    /* Examine status register */
  do {
    if (UsartStatus & US_RXRDY) {          /* Data received?          */
      if (UsartStatus & USART_RX_ERROR_FLAGS) {  /* Any error ?       */
        Dummy = US_RHR;                    /* => Discard data         */
        US_CR = (1uL << 8);                  /* RSTSTA: Reset Status Bits PARE, FRAME, OVRE and RXBRK */
      } else {
        OS_OnRx(US_RHR);                   /* Process actual byte     */
      }
    }
    if (UsartStatus & US_IMR & US_TX_INT_FLAG) { /* Check Tx status => Send next character */
      if (OS_OnTx()) {                     /* No more characters to send ?  */
        US_IDR = US_TX_INT_FLAG;           /* Disable further tx interrupts */
      }
    }
    UsartStatus = US_CSR;                  /* Examine current status  */
  } while (UsartStatus & US_IMR & (US_TX_INT_FLAG | US_RXRDY));
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Never call this function directly from your application
*/
static OS_U8 buffer1;

void OS_COM_Send1(OS_U8 c) {
  buffer1 = c;
  US_TPR  = (OS_U32) &buffer1;
  US_TCR  = 1;
  US_PTCR = (1uL << 8);    /* Enable transmission */
  US_IER = US_TX_INT_FLAG; /* enable Tx interrupt */
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize UART for OSView
*/
void OS_COM_Init(void) {
  OS_IncDI();
  PMC_PCER  = (1uL << OS_UART_ID);    /* Enable peripheral clock for selected USART */
  PIOA_PDR  = PIO_PDR_USART;          /* Enable peripheral output signals (disable PIO Port A) */
  PIOA_ASR  = PIO_PDR_USART;          /* Select "A" peripherals on PIO A (USART Rx, TX) */
  PIOA_BSR  = 0;                      /* Deselect "B" peripherals on PIO A */
  US_CR   = (1uL <<  2) |             /* RSTRX: Reset Receiver: 1 = The receiver logic is reset. */
            (1uL <<  3);              /* RSTTX: Reset Transmitter: 1 = The transmitter logic is reset. */
  US_CR   = (0uL <<  2) |             /* RSTRX: Reset Receiver: 1 = The receiver logic is reset. */
            (0uL <<  3) |             /* RSTTX: Reset Transmitter: 1 = The transmitter logic is reset. */
            (1uL <<  4) |             /* RXEN: Receiver Enable: 1 = The receiver is enabled if RXDIS is 0. */
            (0uL <<  5) |             /* RXDIS: Receiver Disable: 0 = Noeffect. */
            (1uL <<  6) |             /* TXEN: Transmitter Enable: 1 = The transmitter is enabled if TXDIS is 0. */
            (0uL <<  7) |             /* TXDIS: Transmitter Disable: 0 = Noeffect. */
            (1uL <<  8) |             /* RSTSTA: Reset Status Bits: 1 = Resets the status bits PARE, FRAME, OVRE and RXBRK in the US_CSR. */
            (0uL <<  9) |             /* STTBRK: Start Break: 0 = Noeffect. */
            (0uL << 10) |             /* STPBRK: Stop Break: 0 = Noeffect. */
            (0uL << 11) |             /* STTTO: Start Time-out: 0 = Noeffect. */
            (0uL << 12);              /* SENDA: Send Address: 0 = Noeffect. */
  US_MR   = (0uL <<  4) |             /* USCLKS: Clock Selection: 0 = MCK */
            (3uL <<  6) |             /* CHRL: Character Length: 3 = Eight bits */
            (0uL <<  8) |             /* SYNC: Synchronous Mode Select: 0 = USART operates in Asynchronous Mode. */
          (0x4uL <<  9) |             /* PAR: Parity Type: 0x4 = No parity */
            (0uL << 12) |             /* NBSTOP: Number of Stop Bits: 0 = 1 stop bit */
            (0uL << 14) |             /* CHMODE: Channel Mode: 0 = Normal mode */
            (0uL << 17) |             /* MODE9: 9-bit Character Length: 0 = CHRL defines character length. */
            (0uL << 18);              /* CKLO: Clock Output Select: 0 = The USART does not drive the SCK pin. */
  US_BRGR = (OS_PCLK_UART / (OS_BAUDRATE * 16));
  US_IDR  = 0xFFFFFFFF;               /* Disable all interrupts     */
  US_IER  = (1uL << 0) |              /* Enable Rx Interrupt        */
            (0uL << 1);               /* Do not Enable Tx Interrupt */
  /* Setup interrupt controller for OS USART */
  OS_ARM_InstallISRHandler(OS_UART_ID, &OS_COM_ISR_Usart);  /* OS UART interrupt handler vector */
  OS_ARM_ISRSetPrio(OS_UART_ID, OS_UART_PRIO);              /* Level sensitive, selected priority. */
  AIC_ICCR = (1uL << OS_UART_ID);                           /* Clear OS usart interrupt.        */
  OS_ARM_EnableISR(OS_UART_ID);                             /* Enable OS usart interrupts       */
  OS_DecRI();
}

#elif OS_USE_DCC

/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnTx();
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize DCC for OSView
*/
void OS_COM_Init(void) {
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Send 1 character via DCC
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKDCC_SendCharOnChannelNB(1, c);
}

#else  /* DCC not selected, selected UART not supported, using dummies */

void OS_COM_Init(void) {}
void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* avoid compiler warning */
  OS_COM_ClearTxActive();  /* let the OS know that Tx is not busy */
}

#endif /*  OS_UART_USED  */

/****** Final check of configuration ********************************/

#ifndef OS_UART_USED
  #error "OS_UART_USED has to be defined"
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler
*
*       Detect reason for IRQ and call corresponding service routine.
*       OS_irq_handler is called from OS_IRQ_SERVICE function
*       found in RTOSVect.asm
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  pISR = (OS_ISR_HANDLER*) AIC_IVR;    // Read interrupt vector to release NIRQ to CPU core
#if DEBUG
  AIC_IVR = (OS_U32) pISR;             // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
#if ALLOW_NESTED_INTERRUPTS
  OS_EnterNestableInterrupt();         // Now interrupts may be reenabled. If nesting should be allowed
#else
  OS_EnterInterrupt();                 // Inform OS that interrupt handler is running
#endif
  pISR();                              // Call interrupt service routine
  OS_DI();                             // Disable interrupts and unlock
  AIC_EOICR = 0;                       // interrupt controller =>  Restore previous priority
#if ALLOW_NESTED_INTERRUPTS
  OS_LeaveNestableInterrupt();         // Leave nestable interrupt, perform task switch if required
#else
  OS_LeaveInterrupt();                 // Leave interrupt, perform task switch if required
#endif
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return NULL;
  }
#endif
  OS_DI();
  papISR = (OS_ISR_HANDLER**)AIC_SVR_BASE_ADDR;
  pOldHandler          = *(papISR + ISRIndex);
  *(papISR + ISRIndex) = pISRHandler;
  OS_RestoreI();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  AIC_IECR = (1uL << ISRIndex);
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_DisableISR
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  AIC_IDCR = (1uL << ISRIndex);
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32* pPrio;
  int     OldPrio;

#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return 0;
  }
#endif
  OS_DI();
  pPrio = (OS_U32*)AIC_SMR_BASE_ADDR;
  OldPrio = pPrio[ISRIndex];
  pPrio[ISRIndex] = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
  OS_RestoreI();
  return OldPrio & INT_PRIORITY_MASK;
}

/*********************************************************************
*
*       __low_level_init()
*
*       Initialize memory controller, clock generation and pll
*
*       Has to be modified, if another CPU clock frequency should be
*       used. This function is called during startup and
*       has to return 1 to perform segment initialization
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
//
// Calculation of Flash microsecond cycle numbers for flash programming
// MCLK frequency (in MHz) * 1.5. The value has to be rounded up
//
#define FMCN_VAL ((((OS_FSYS + 500000) / 1000000) * 3 ) >> 1)

OS_INTERWORK int __low_level_init(void) {
  WDT_MR = (1uL << 15);        // Initially disable watchdog
  //
  // Setup Flash microsecond cycle numbers and wait states
  //
  MC_FMR0 = ((0xFF << 16) & (FMCN_VAL << 16)) | MC_FWS;
  if ( ((DBGU_CPUID >> 8) & 0xFF) >= 0x09) {
    //
    // The Flash size is above 256KB, initialize the second Flash controller
    //
    MC_FMR1 = ((0xFF << 16) & (FMCN_VAL << 16)) | MC_FWS;
  }
  //
  // Start-up main oscillator
  //
  PMC_MOR  = (1uL << 0)       // Enable main oscillator
            | (0x40 << 8)     // Set startup delay time, 512 slow clock cycles
            ;
  //
  // Wait until end of startup time => MOSCS set to 1
  //
  while ((PMC_SR & (1uL << 0)) == 0);
  //
  // Initialize PLL to run at 96 MHz
  //
  PMC_PLLR = 0x05             // Divider = 5
            | (28  << 8)      //
            | (25  << 16)     //
            | (1uL << 28)     // USBDIV is value + 1 = 2
            ;
  //
  // Wait until pll is locked, until LOCK bit is set to 1
  //
  while((PMC_SR & (1uL << 2)) == 0);
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Select main clock as master clock, set divider
  //
  PMC_MCKR = 0                // Select main clock oscillator as source
            | (1uL <<  2)     // Divide clock by 2
            ;
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Select PLL clock as master clock, set divider
  //
  PMC_MCKR = 3                // Select PLL clock as source
            | (1uL <<  2)     // Divide clock by 2
            ;
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Initialize interrupt controller
  //
  _InitVIC();                 // Initialize interrupt controller, clear and reset all interrupts
  //
  // Initialize real time timer
  //
  _InitRTT();
  //
  // Initialize reset control
  //
  RSTC_MR = (((OS_U32)0xA5 << 24) | (1uL << 0));  // write KEY and URSTEN to allow USER RESET
  return 1;
}

/*********************************************************************
*
*       Optional speed optimization for IAR
*
**********************************************************************
*/
#ifdef __ICCARM__

#ifndef   OPTIMIZE_SPEED
  #define OPTIMIZE_SPEED (0)
#endif

#if OPTIMIZE_SPEED
/*********************************************************************
*
*       OS_EnableInt()
*/
OS_INTERWORK __arm __ramfunc void OS_EnableInt(void) {
  __set_CPSR(__get_CPSR() & ~0x80uL); // Clear IRQ only, keep other bits unchanged
}

/*********************************************************************
*
*       OS_DisableInt()
*/
OS_INTERWORK __arm __ramfunc void OS_DisableInt(void) {
  __set_CPSR(__get_CPSR() | 0x80); // Set IRQ only, keep other bits unchanged
}
#endif  // OPTIMIZE_SPEED
#endif  // __ICCARM__

/****** End Of File *************************************************/
