/*********************************************************************
*                SEGGER Microcontroller GmbH & Co. KG                *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2016 SEGGER Microcontroller GmbH & Co. KG         *
*                                                                    *
*     Internet: segger.com   Support: support_embos@segger.com       *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.16                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_R7S72100.c for the Renesas RZ1 CortexA9 CPU

Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.

Literature:
  [1]  Generic Interrupt Controller (GIC-390) Errata Notice
       (\\fileserver\Techinfo\Company\ARM\GIC_GenericInterruptController\GIC_390_Errata_Notice_v6.pdf)
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "SEGGER_SYSVIEW.h"

#ifdef __ICCARM__
  #include <intrinsics.h>
#endif

#ifdef __GNUC__
  //
  //  Suppress warning about set but unused variable "Dummy"
  //
  #pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#endif

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

#define ALLOW_NESTED_INTERRUPTS 0   // Caution: Nesting interrupts will cause higher stack load on system stack CSTACK

#ifndef   COPY_TO_SRAM              // Code and data may be copied from external Flash into internal SRAM
  #define COPY_TO_SRAM (0)
#endif

/*********************************************************************
*
*       Clock frequency settings
*/
#ifndef   OS_FSYS                   /* CPU Main clock frequency     */
  #define OS_FSYS (399900000u)      /* Assume 10MHz                 */
#endif

#ifndef   OS_PCLK_TIMER             /* Peripheral clock for timer   */
  #define OS_PCLK_TIMER (OS_FSYS/12)/* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_PCLK_UART              /* Peripheral clock for UART    */
  #define OS_PCLK_UART (OS_FSYS/6)  /* May vary from CPU clock      */
#endif                              /* depending on CPU             */

#ifndef   OS_TICK_FREQ
  #define OS_TICK_FREQ (1000)
#endif

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_ENABLE            // Global enable of communication
  #define OS_VIEW_ENABLE   (1)      // Default: on
#endif

#ifndef   OS_VIEW_USE_UART          // If set, UART will be used for communication
  #define OS_VIEW_USE_UART (0)      // Default: 0 => DCC is used
#endif                              // if OS_VIEW_ENABLE is on

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       dedicate a UART to embOSView, please define it to be -1
*       Currently UART2 is supported and the standard
*       setup enables UART 2 per default.
*       On RSKRZA1 the UART2 is available via the CN18 USB connector.
*       When using DCC for communication, the UART is not used for embOSView,
*       regardless the OS_UART settings
*/
#ifndef   OS_UART
  #define OS_UART (2)
#endif

#ifndef   OS_BAUDRATE
  #define OS_BAUDRATE (38400)
#endif

/****** End of configurable options *********************************/

/*********************************************************************
*
*       DCC and UART settings for embOSView
*
*       Automatically generated from configuration settings
*/
#define OS_USE_DCC     (OS_VIEW_ENABLE && (OS_VIEW_USE_UART == 0))
#define OS_UART_USED   (OS_VIEW_ENABLE && (OS_VIEW_USE_UART != 0)) && (OS_UART == 2)

#if (OS_UART_USED || OS_USE_DCC)
  #define OS_COM_INIT() OS_COM_Init()
#else
  #define OS_COM_INIT()
#endif

#if ((defined(__ICCARM__) && DEBUG) || OS_USE_DCC)
  #include "JLINKDCC.h"
#endif

/********************************************************************/

#ifdef __ICCARM__
  #define OS_ARM_FUNC  __arm
  #define ATTRIBUTE_NAKED_NO_INLINE
#endif

#ifdef __GNUC__
  #define OS_ARM_FUNC
  #define ATTRIBUTE_NAKED_NO_INLINE __attribute__ ((naked, noinline))
  #pragma GCC diagnostic ignored "-Wunused-parameter" // Avoid warning for unused parameter with GCC
#endif

#ifdef __CC_ARM
  #define OS_ARM_FUNC
  #define OS_INTERWORK
  #define ATTRIBUTE_NAKED_NO_INLINE __attribute__ ((noinline))
#endif

/*********************************************************************
*
*       Check configuration
*
**********************************************************************
*/
#ifndef   DEBUG     /* Should normally be defined as project option */
  #define DEBUG  (0)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
/* System, clock and standby control */
#define CPG_BASE_ADDR                (0xFCFE0000)
#define CPG_FRQCR                    (*(volatile unsigned short*) (CPG_BASE_ADDR + 0x0010u))
#define CPG_FRQCR2                   (*(volatile unsigned short*) (CPG_BASE_ADDR + 0x0014u))
#define CPG_SYSCR3                   (*(volatile unsigned char*)  (CPG_BASE_ADDR + 0x0408u))
#define CPG_STBCR4                   (*(volatile unsigned char*)  (CPG_BASE_ADDR + 0x0424u))
#define CPG_STBCR5                   (*(volatile unsigned char*)  (CPG_BASE_ADDR + 0x0428u))

/* BUS controller */
#define BSC_BASE_ADDR                (0x3FFFC000u)
#define BSC_CS0BCR                   (*(volatile unsigned long*)  (BSC_BASE_ADDR + 0x0004u))
#define BSC_CS0WCR                   (*(volatile unsigned long*)  (BSC_BASE_ADDR + 0x0028u))

/* OS timer */
#define OSTM0_BASE_ADDR              (0xFCFEC000u)
#define OSTM1_BASE_ADDR              (0xFCFEC400u)
#define OSTM0_INT_ID                 (134)
#define OSTM1_INT_ID                 (135)
#define OSTM0_STB_BIT                (1)
#define OSTM1_STB_BIT                (0)
#define OSTM_BASE_ADDR               (OSTM0_BASE_ADDR)  /* Select the timer used for OS */
#define OSTM_INT_ID                  (OSTM0_INT_ID)
#define OSTM_STB_BIT                 (OSTM0_STB_BIT)
#define OSTM_CMP                     (*(volatile unsigned int*)  (OSTM_BASE_ADDR + 0x00u))
#define OSTM_CNT                     (*(volatile unsigned int*)  (OSTM_BASE_ADDR + 0x04u))
#define OSTM_TE                      (*(volatile unsigned char*) (OSTM_BASE_ADDR + 0x10u))
#define OSTM_TS                      (*(volatile unsigned char*) (OSTM_BASE_ADDR + 0x14u))
#define OSTM_TT                      (*(volatile unsigned char*) (OSTM_BASE_ADDR + 0x18u))
#define OSTM_CTL                     (*(volatile unsigned char*) (OSTM_BASE_ADDR + 0x20u))

/* Watchdog */
#define WT_BASE_ADDR                 (0xFCFE0000u)
#define WT_CSR                       (*(volatile unsigned short*)  (WT_BASE_ADDR + 0x00u))
#define WT_CNT                       (*(volatile unsigned short*)  (WT_BASE_ADDR + 0x02u))
#define WR_CSR                       (*(volatile unsigned short*)  (WT_BASE_ADDR + 0x04u))

/* Interrupt controller GIC */
#define GIC_BASE_ADDR                (0xE8200000u)
#define GICD_BASE_ADDR               (GIC_BASE_ADDR + 0x1000u)
#define GICC_BASE_ADDR               (GIC_BASE_ADDR + 0x2000u)
#define GICC_HPPIR                   (*(volatile unsigned int*) (GICC_BASE_ADDR + 0x018u))  // Highest Priority Pending Interrupt Register
#define GICD_IPRIORITY               (*(volatile unsigned int*) (GICD_BASE_ADDR + 0x400u))  // Interrupt Priority Registers

#define NUM_INT_SOURCES              (587u)  /* The Renesas RZ1 R7S72100 has 587 interrupt sources */

/* L2 cache */
#define L2CACHE_BASE_ADDR            0x3FFFF000u

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
extern const OS_ARM_L2CACHE_API OS_L2CACHE_L2C310;

extern unsigned int __vector;             // Start of vector table, defined in startup file

/*********************************************************************
*
*       MMU and cache configuration
*
* NOTES:
*   The translation table has to be aligned on 16384 bytes boundary
*   The translation table has to be located in un-initialized memory,
*   because it may be filled from startup, before segments and variables
*   are initialized.
*/
#ifdef __ICCARM__
  #pragma data_alignment=16384
  static __no_init unsigned int _TranslationTable [0x1000];
#elif defined __CC_ARM
  #pragma arm section zidata = "NO_INIT"         //   Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static unsigned int _TranslationTable[0x1000];
  #pragma arm section zidata // back to default (.bss section)
#elif defined __GNUC__
  static unsigned int _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".no_init")));
#endif

/*********************************************************************
*
*       Interrupt vector table
*
* NOTES:
*   The translation table has to be located in un-initialized memory,
*   because it may be filled from startup, before segments and variables
*   are initialized.
*/
#ifdef __ICCARM__
  #pragma data_alignment=4
  static __no_init OS_ISR_HANDLER * _apIsrHandler[NUM_INT_SOURCES];
#elif defined __CC_ARM
  #pragma arm section zidata = "NO_INIT"         //   Has to be declared as separate section UNINIT in the scatter file
  __align(4) static OS_ISR_HANDLER * _apIsrHandler[NUM_INT_SOURCES];
  #pragma arm section zidata // back to default (.bss section)
#elif defined __GNUC__
  static OS_ISR_HANDLER * _apIsrHandler[NUM_INT_SOURCES] __attribute__ ((aligned (4), section (".no_init")));
#endif

/****** Interrupt configuration registers default table.         ****/
static const OS_U32 _INTC_ICDICFRN_Table[] = {
  0xAAAAAAAA,            /* ICDICFR0  */
  0x00000055,            /* ICDICFR1  */
  0xFFFD5555,            /* ICDICFR2  */
  0x555FFFFF,            /* ICDICFR3  */
  0x55555555,            /* ICDICFR4  */
  0x55555555,            /* ICDICFR5  */
  0x55555555,            /* ICDICFR6  */
  0x55555555,            /* ICDICFR7  */
  0x5555F555,            /* ICDICFR8  */
  0x55555555,            /* ICDICFR9  */
  0x55555555,            /* ICDICFR10 */
  0xF5555555,            /* ICDICFR11 */
  0xF555F555,            /* ICDICFR12 */
  0x5555F555,            /* ICDICFR13 */
  0x55555555,            /* ICDICFR14 */
  0x55555555,            /* ICDICFR15 */
  0x55555555,            /* ICDICFR16 */
  0xFD555555,            /* ICDICFR17 */
  0x55555557,            /* ICDICFR18 */
  0x55555555,            /* ICDICFR19 */
  0xFFD55555,            /* ICDICFR20 */
  0x5F55557F,            /* ICDICFR21 */
  0xFD55555F,            /* ICDICFR22 */
  0x55555557,            /* ICDICFR23 */
  0x55555555,            /* ICDICFR24 */
  0x55555555,            /* ICDICFR25 */
  0x55555555,            /* ICDICFR26 */
  0x55555555,            /* ICDICFR27 */
  0x55555555,            /* ICDICFR28 */
  0x55555555,            /* ICDICFR29 */
  0x55555555,            /* ICDICFR30 */
  0x55555555,            /* ICDICFR31 */
  0x55555555,            /* ICDICFR32 */
  0x55555555,            /* ICDICFR33 */
  0x55555555,            /* ICDICFR34 */
  0x55555555,            /* ICDICFR35 */
  0x55555555             /* ICDICFR36 */
};

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_ISR_Undefined
*
*       Is called when an uninstalled interrupt was detected
*       As interrupt pending condition of peripherals has to be reset,
*       program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(void);
static void _OS_ISR_Undefined(void) {
  volatile int Dummy;
  Dummy = 1;
  /* You may set a breakpoint here to detect uninstalled ISR-handler */
  while (Dummy > 0) {
  }
}

/*********************************************************************
*
*       _SpuriousIrqHandler(), a dummy spurious IRQ handler
*/
static OS_U32 _SpuriousIrqCnt;
static void _SpuriousIrqHandler(void) {
  _SpuriousIrqCnt++;
}

/*********************************************************************
*
*       OS_ISR_Tick()
*       The OS system timer interrupt
*/
static void OS_ISR_Tick (void);
static void OS_ISR_Tick (void) {
  OS_TICK_Handle();

  #if ((defined(__ICCARM__) && DEBUG) || OS_USE_DCC)
    JLINKDCC_Process();
  #endif
}

/*********************************************************************
*
*       _InitPLL()
*
*       Initialize PLL to speed up system
*       This function is called from __low_level_init() before
*       any variables are initialized.
*       Therefore, it can not rely on any variable content
*       and can not set any variables (except no-init variables).
*/
static void _InitPLL(void) {
  volatile OS_U32 Dummy;

  *(volatile OS_U32 *)(0x3fffff80u) = 0x00000001u;  // Magic access. Address not described in Manual
  Dummy = *(volatile OS_U32 *)(0x3fffff80u);
  //
  // Set up PLL, assuming 13.33 MHz external crystal, Multiply by 30
  //
  CPG_FRQCR   = (0x35u <<  0)  // b7..b0:   Reserved, always 0x35
              | (0x0   <<  8)  // b9..b8:   CPU clock division ratio. 0 => 1/1
              | (0x0   << 10)  // b11..b10: Reserved, always 0
              | (0x1u  << 12)  // b13..b12: Clock output enable: low level in Standby and Deep standby mode
              | (0x0   << 14)  // b14:      Unstable clock output when XTAL gain changes
              ;                // b15 always 0

  CPG_FRQCR2  = 0x0003u;       // Image Processing Clock division ratio. 3 => 1/3
  //
  // Enable on chip data retention RAM
  //
  CPG_SYSCR3  = 0x0Fu;
}

/*********************************************************************
*
*       _GIC_Init()
*
*       Initialize interrupt controller by setting default vectors
*       and clearing all interrupts
*/
static void _GIC_Init(void) {
  OS_U32 i;
  //
  // Set GIC base address
  //
  OS_GIC_SetBaseAddr(0xE8200000u);
  //
  // Configure all interrupts
  //
  for (i = 0; i < NUM_INT_SOURCES; i++) {
    _apIsrHandler[i] = _OS_ISR_Undefined;  // Set default interrupt handler

    OS_GIC_SetIntSecure(i);                // Set to secure
    OS_GIC_DisableInt(i);                  // Disable interrupts
    //
    // Set interrupts to level or edge sensitive
    //
    OS_GIC_SetIntCfg(i, (((OS_U8*)_INTC_ICDICFRN_Table)[i / 4] >> ((i * 2u) % 8u)) & 0x03u);
    //
    // Initialize all Interrupt pointer target register to address the only A9 core
    // The first 8 registers are fixed
    //
    if (i > 8u) {
      OS_GIC_SetTargetProc(i, 0x01);
    }
  }
  //
  // Configure priority mask to lowest value (highest number) - enables all priorities
  //
  OS_GIC_SetPrioThreshold(0xFF);
  //
  // This register defines the point at which the priority value fields split into two parts,
  // the group priority field and the subpriority field.
  //
  OS_GIC_SetPrioGroup(0x00);
  //
  // Enable the GIC
  //
  OS_GIC_Enable();
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*   Returns the current hardware timer count value
*
* Return value
*   Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return OSTM_CNT;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*   Returns if the hardware timer interrupt pending flag is set
*
* Return value
*   == 0; Interrupt pending flag not set
*   != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (OS_ARM_IsPending(OSTM_INT_ID));
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
#define OS_TIMER_RELOAD ((OS_PCLK_TIMER/OS_TICK_FREQ) - 1)
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_PCLK_TIMER, OS_TICK_FREQ, 0u, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_U32 TickPrio;
  volatile OS_U32 Dummy;

  OS_IncDI();
  OS_ARM_CACHE_Sync();                                 // Ensure, caches are synchronized
  //
  // Set vector table base address
  //
  OS_ARM_SetVBAR((OS_U32)&__vector);
  //
  // Initialize GIC
  //
  _GIC_Init();
  //
  // Install OS timer ISR Handler and set interrupt priority to lowest priority
  //
  OS_ARM_InstallISRHandler(OSTM_INT_ID, OS_ISR_Tick);
  OS_ARM_ISRSetPrio(OSTM_INT_ID, 0xFFu);               // Set to lowest level, ALL BITS set
  TickPrio  = OS_ARM_ISRSetPrio(OSTM_INT_ID, 0xFFu);   // Read back priority to examine relevant preemption-level-bits
  TickPrio -= 1;                                       // Set to low preemption level, 1 level higher than lowest
  OS_ARM_ISRSetPrio(OSTM_INT_ID, TickPrio);
  OS_ARM_ClearPendingFlag(OSTM_INT_ID);
  OS_ARM_EnableISR(OSTM_INT_ID);
  //
  // Initialize OS timer
  //
  CPG_STBCR5 &= ~(1u << OSTM_STB_BIT);  // Release OS timer standby
  Dummy       = CPG_STBCR5;             // Dummy read (may not be required)
  OSTM_TT     = 0x01u;                  // Stop Timer
  OSTM_CTL    = (0u << 0)               // b0: 0 => Do not enable interrupt on start
              | (0u << 1);              // b1: 0 => Interval timer mode
                                        // All other bits 0
  OSTM_CMP    = OS_TIMER_RELOAD;        // Set timer interval
  OSTM_TS     = 0x01u;                  // Start timer

  //
  // Setup values for usec precise system time functions
  //
  OS_Config_SysTimer(&SysTimerConfig);
  //
  // Configure and initialize SEGGER SystemView
  //
#if OS_PROFILE
  SEGGER_SYSVIEW_Conf();
#endif
  //
  // Initialize the optional communication for embOSView
  //
  OS_COM_INIT();
  OS_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*       Please note:
*       This is basically the "core" of the idle loop.
*       This core loop can be changed, but:
*       The idle loop does not have a stack of its own, therefore no
*       functionality should be implemented that relies on the stack
*       to be preserved. However, a simple program loop can be programmed
*       (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if (DEBUG == 0)
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*       This routine is required for task-info via OSView or high
*       resolution time measurement functions.
*       It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_U32 Cnt;

  Time = OS_GetTime32();
  Cnt  = OS_TIMER_RELOAD - OSTM_CNT;
  //
  // Check if timer interrupt pending ...
  //
  if (OS_ARM_IsPending(OSTM_INT_ID) != 0) {
    Cnt = OS_TIMER_RELOAD - OSTM_CNT;     // Interrupt pending, re-read timer and adjust result
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*       Convert Cycles into micro seconds.
*
*       If your clock frequency is not a multiple of 1 MHz,
*       you may have to modify this routine in order to get proper
*       diagnostics.
*
*       This routine is required for profiling or high resolution time
*       measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
#define OS_TIMER_FREQ (OS_PCLK_TIMER)

#if (OS_TIMER_FREQ/1000000 * 1000000 == OS_TIMER_FREQ)
  return Cycles/(OS_PCLK_TIMER/1000000);
#else
  long long t;
  t = Cycles;
  t *= 1000000;
  t /= OS_TIMER_FREQ;
  return (OS_U32)t;
#endif
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if OS_UART_USED

#define OS_UART_PRIO         (0xF0)        /* Lowest priority for UART interrupts */
#define PORT3_BASE_ADDR      (0xFCFE3000)

//
// Assign UART to port and peripheral
//
#define UART_PORT_BASE_ADDR  (PORT3_BASE_ADDR)
#define UART_TX_BIT          (0)
#define UART_RX_BIT          (2)

#define UART_PORT_PMC        (*(volatile unsigned short*) (UART_PORT_BASE_ADDR + 0x040Cu))
#define UART_PORT_PFCAE      (*(volatile unsigned short*) (UART_PORT_BASE_ADDR + 0x0A0Cu))
#define UART_PORT_PFCE       (*(volatile unsigned short*) (UART_PORT_BASE_ADDR + 0x060Cu))
#define UART_PORT_PFC        (*(volatile unsigned short*) (UART_PORT_BASE_ADDR + 0x050Cu))
#define UART_PORT_PIPC       (*(volatile unsigned short*) (UART_PORT_BASE_ADDR + 0x420Cu))

#define SCFI2_BASE_ADDR      (0xE8008000)
#define UART_BASE_ADDR       (SCFI2_BASE_ADDR)

#define OS_UART_ERR_INT_ID   (230)
#define OS_UART_RX_INT_ID    (231)
#define OS_UART_TX_INT_ID    (232)

#define UART_SCSMR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0000u))
#define UART_SCBRR           (*(volatile unsigned char*)  (UART_BASE_ADDR + 0x0004u))
#define UART_SCSCR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0008u))
#define UART_SCFTDR          (*(volatile unsigned char*)  (UART_BASE_ADDR + 0x000Cu))
#define UART_SCFSR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0010u))
#define UART_SCFRDR          (*(volatile unsigned char*)  (UART_BASE_ADDR + 0x0014u))
#define UART_SCFCR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0018u))
#define UART_SCSPTR          (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0020u))
#define UART_SCLSR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0024u))
#define UART_SCEMR           (*(volatile unsigned short*) (UART_BASE_ADDR + 0x0028u))

#define BRG_VALUE            ((OS_PCLK_UART+OS_BAUDRATE*16L)/(OS_BAUDRATE*32L)-1)

#if (BRG_VALUE < 35u)
  #error "Baudrate generator value may produce an error rate above 1 percent, should be verified"
#endif
#if (BRG_VALUE > 255u)
  #error "Baudrate generator value too high. Prescaler has to be modified"
#endif

#define UART_FE_ERROR_BIT    (3)
#define UART_PE_ERROR_BIT    (2)
#define UART_RX_ERROR_BIT    (7)
#define UART_RDF_BIT         (1)
#define UART_RXRDY_BIT       (0)
#define UART_TEND_BIT        (6)
#define UART_TDFE_BIT        (5)
#define UART_RX_ERROR_FLAGS  (1u << UART_RX_ERROR_BIT)
#define UART_RXRDY           (1u << UART_RDF_BIT)
#define UART_TXRDY           (1u << UART_TDFE_BIT)

/*********************************************************************
*
*       _COM_ISR_Tx()
*       OS UART Tx interrupt handler
*/
static void _COM_ISR_Tx(void) {
  unsigned int UsartStatus;

  UsartStatus = UART_SCFSR;       // Examine status register
  if (UsartStatus & UART_TXRDY) { // Check Tx status => Send next character
    //
    // Clear TX interrupt state
    //
    UART_SCSCR &= ~(1u << 7);     // Disable Tx interrupt
    UART_SCFSR &= ~(UART_TXRDY);
    OS_OnTx();                    // Send next charaters (if any)
  }
}

/*********************************************************************
*
*       _COM_ISR_Rx()
*       OS UART Rx interrupt handler
*/
static void _COM_ISR_Rx(void) {
  unsigned int UsartStatus;
  unsigned int Data;

  UsartStatus = UART_SCFSR;                   /* Examine status register */
  if (UsartStatus & UART_RXRDY) {             /* Data received?          */
    Data = UART_SCFRDR;
    UART_SCFSR &= ~(UART_RXRDY);              /* Clear Rx ready flag, reset interrupt pending */
    if (UsartStatus & UART_RX_ERROR_FLAGS) {  /* Any error ?         */
      UART_SCFSR &= ~(UART_RX_ERROR_FLAGS);   /* Clear error flags   */
    } else {
      OS_OnRx(Data);                          /* Process actual byte */
    }
  }
}

/*********************************************************************
*
*       _COM_ISR_RxError()
*       OS UART Rx error interrupt handler
*/
static void _COM_ISR_RxError(void) {
  unsigned int          UsartStatus;
  volatile unsigned int Dummy;

  UsartStatus = UART_SCFSR;                 /* Examine status register */
  if (UsartStatus & UART_RX_ERROR_FLAGS) {  /* Any error ?             */
    Dummy = UART_SCFRDR;                    /* => Discard data         */
    UART_SCFSR &= ~(UART_RX_ERROR_FLAGS);   /* Clear Error Flags       */
  }
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Send one character via UART.
*       Never call this function from your application
*/
void OS_COM_Send1(OS_U8 c) {
  UART_SCFTDR = c;
  UART_SCSCR |= (1u << 7);   // b7: Enable Tx interrupt
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize UART for embOSView (supports SCIF channel 2)
*/
void OS_COM_Init(void) {
  OS_IncDI();
  //
  // Initially, the clock for UART module has to be enabled
  //
  CPG_STBCR4 &= ~(1u << 5);  // Enable clock for SCIF channel 2
  //
  // Now initialize the UART module
  //
  UART_SCSCR  = 0x0000u;    // Stop UART, select internal clock
  UART_SCFCR  = (0u << 0)   // Loopback mode disabled
              | (1u << 1)   // Reset Receive FIFO
              | (1u << 2)   // Reset Transmit FIFO
              | (0u << 3)   // Disable Modem control signals
              | (3u << 4)   // b5..b4: Transmit FIFO data trigger: 0 bytes, disable FIFO
              | (0u << 6)   // b7..b6;  Receive FIFO date trigger, 1 byte
              | (0u << 8)   // b10..b8: RTS active trigger, don't care, not used
              ;             // All other bits reserved, always 0

  UART_SCFSR &= 0xFF6E;     // Reset error flags (Has to be Read AND Write operation)
  UART_SCLSR &= 0x0000;     // Clear Overrun error, Has to be Read Modify write
  UART_SCSMR  = (0u << 0)   // b1..b0: Clock source divider. 0 => 1/1; 1 => 1/4; 2 => 1/16; 3 => 1/64
              | (0u << 2)   // b2:     Reserved, always 0
              | (0u << 3)   // b3:     1 Stop bit
              | (0u << 4)   // b4:     EVEN parity
              | (0u << 5)   // b5:     Parity disabled
              | (0u << 6)   // b6:     8 data bits
              | (0u << 7)   // b7:     Mode, async
              ;             // All others reserved, set to 0

  UART_SCEMR  = (0u << 0)   // Async base clock oversampling 1/16
              | (0u << 7)   // Baud rate generator normal speed mode
              ;             // All other bits reserved, always 0

  UART_SCBRR = BRG_VALUE;   // Set baudrate
  UART_SCFCR  = (0u << 0)   // Loopback mode disabled
              | (0u << 1)   // Release Reset for Receive FIFO
              | (0u << 2)   // Release Reset for Transmit FIFO
              | (0u << 3)   // Disable Modem control signals
              | (3u << 4)   // b5..b4: Transmit FIFO data trigger: 0 bytes, disable FIFO
              | (0u << 6)   // b7..b6;  Receive FIFO date trigger, 1 byte
              | (0u << 8)   // b10..b8: RTS active trigger, don't care, not used
              ;             // All other bits reserved, always 0

  UART_SCSPTR = (1u << 0)   // Break output, output is high level
              | (1u << 0)   // Break bit is passed to TxD output
              ;             // Handshake not used, all other bits 0
  //
  // Setup port for TxD
  //
  UART_PORT_PMC     |=  (1u << UART_TX_BIT);  // Alternate function TxD
  UART_PORT_PFCAE   |=  (1u << UART_TX_BIT);  // Alternate function 6
  UART_PORT_PFCE    &= ~(1u << UART_TX_BIT);  // Alternate function 6
  UART_PORT_PFC     |=  (1u << UART_TX_BIT);  // Alternate function 6
  UART_PORT_PIPC    |=  (1u << UART_TX_BIT);  // Datat direction controlled by alternate function
  //
  // Setup port for RxD
  //
  UART_PORT_PMC     |=  (1u << UART_RX_BIT);  // Alternate function RxD
  UART_PORT_PFCAE   &= ~(1u << UART_RX_BIT);  // Alternate function 4
  UART_PORT_PFCE    |=  (1u << UART_RX_BIT);  // Alternate function 4
  UART_PORT_PFC     |=  (1u << UART_RX_BIT);  // Alternate function 4
  UART_PORT_PIPC    |=  (1u << UART_RX_BIT);  // Datat direction controlled by alternate function
  //
  // Now enable Rx Tx and Interrupts
  //
  UART_SCSCR  = (0u << 0)   // b1..0: Clock select, Internal clock used
              | (0u << 2)   // b2:    Reserved, always 0
              | (0u << 3)   // b3:    Receive error interrupt, disabled
              | (1u << 4)   // b4:    Rx enabled
              | (1u << 5)   // b5:    Tx enabled
              | (1u << 6)   // b6:    Rx interrupt enabled
              | (0u << 7)   // b7:    Tx interrupt disabled per default
              ;             // b15..b8 all reserved, always 0
  //
  // Install interrupt handler and enable interrupts
  //
  OS_ARM_InstallISRHandler(OS_UART_RX_INT_ID, _COM_ISR_Rx);
  OS_ARM_ISRSetPrio       (OS_UART_RX_INT_ID, OS_UART_PRIO);
  OS_ARM_ClearPendingFlag (OS_UART_RX_INT_ID);
  OS_ARM_EnableISR        (OS_UART_RX_INT_ID);
  OS_ARM_InstallISRHandler(OS_UART_TX_INT_ID, _COM_ISR_Tx);
  OS_ARM_ISRSetPrio       (OS_UART_TX_INT_ID, OS_UART_PRIO);
  OS_ARM_ClearPendingFlag (OS_UART_TX_INT_ID);
  OS_ARM_EnableISR        (OS_UART_TX_INT_ID);
  OS_ARM_InstallISRHandler(OS_UART_ERR_INT_ID, _COM_ISR_RxError);
  OS_ARM_ISRSetPrio       (OS_UART_ERR_INT_ID, OS_UART_PRIO);
  OS_ARM_ClearPendingFlag (OS_UART_ERR_INT_ID);
  OS_ARM_EnableISR        (OS_UART_ERR_INT_ID);

  OS_DecRI();
}

#elif OS_USE_DCC

/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  (void) Channel;   // Not supported by this version, suppress warning
  OS_OnTx();
}

/*********************************************************************
*
*       OS_COM_Init()
*       Initialize DCC for OSView
*/
void OS_COM_Init(void) {
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
}

/*********************************************************************
*
*       OS_COM_Send1()
*       Send 1 character via DCC
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKDCC_SendCharOnChannelNB(1, c);
}

#else  /* UART for communication not used, define dummy functions */

void OS_COM_Send1(OS_U8 c) {
  OS_USEPARA(c);           /* Avoid compiler warning */
  OS_COM_ClearTxActive();  /* Let the OS know that Tx is not busy */
}
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*       Detect reason for IRQ and call corresponding service routine.
*       OS_irq_handler is called from the IRQ_handler function
*       found in RTOS.s.
*       IRQ_Handler has to be inserted in the vector table in the
*       startup code.
*/
OS_U32 SEGGER_SYSVIEW_TickCnt;
OS_U32 SEGGER_SYSVIEW_InterruptId;

OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER  *pISR;
  volatile OS_U32 Dummy;
  OS_U32          int_id;
  OS_U32          icdipr0_value;

  Dummy  = GICC_HPPIR;       // Erratum 801120 [1] - ensure subsequent ICCIAR data will be correct
  int_id = OS_GIC_AckInt();  // Acknowledge the interrupt and read the interrupt ID

#if OS_PROFILE
  if (int_id == OSTM0_INT_ID) {
	SEGGER_SYSVIEW_TickCnt++;           // Early increment of TickCnt is required by SystemView
  }
  SEGGER_SYSVIEW_InterruptId = int_id;  // Save current interrupt ID to display in SystemView
#endif

  //
  // Read data of 0x0 (SGI ID0 from CPU0) is possibly affected by 733075
  // Interrupt IDs 0x3FE and 0x3FF are also possibly affected
  //
  if ((int_id == 0u) || (int_id  >= 0x3FEu)) {
    //
    // Erratum 733075 - prevent possible CPU interface deadlock (ARM Errata Notice, PR258-GENC-008719 v6.0).
    // Assumes SGI ID0 from CPU0 is not used as suggested by ARM.
    //
    icdipr0_value  = GICD_IPRIORITY; // One way to get the value for the workaround write is to read it first
    GICD_IPRIORITY = icdipr0_value;
    __asm("dsb");                   // Ensure write completes before any future ICCHPIR or ICCIAR reads
  }

  if ((int_id == 0x3FEu) || (int_id == 0x3FFu)) {  // Check for spurious interrupts
     _SpuriousIrqHandler();
  } else {
    if (int_id < NUM_INT_SOURCES) {
      pISR = _apIsrHandler[int_id];      // Get interrupt vector handler
      #if ALLOW_NESTED_INTERRUPTS
      OS_EnterNestableInterrupt();       // Now interrupts may be reenabled. If nesting should be allowed
      #else
      OS_EnterInterrupt();               // Inform OS that interrupt handler is running
      #endif
      //
      // Call interrupt service routine
      //
      pISR();
      #if ALLOW_NESTED_INTERRUPTS
      OS_LeaveNestableInterrupt();       // Leave nestable interrupt, perform task switch if required
      #else
      OS_LeaveInterrupt();               // Leave interrupt, perform task switch if required
      #endif
    }
    OS_GIC_EOI(int_id);                  // End interrupt
  }
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER *pOldHandler;

#if OS_DEBUG
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return NULL;
  }
#endif
  OS_DI();
  pOldHandler             = _apIsrHandler[ISRIndex];
  _apIsrHandler[ISRIndex] = pISRHandler;
  OS_RestoreI();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  OS_GIC_EnableInt(ISRIndex);
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_DisableISR
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  OS_GIC_DisableInt(ISRIndex);
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32  OldPrio;
#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return 0;
  }
#endif
  OS_DI();
  OldPrio = OS_GIC_SetPrio(ISRIndex, Prio);
  OS_RestoreI();
  return (int)OldPrio;
}

/*********************************************************************
*
*       OS_ARM_ClearPendingFlag()
*/
void OS_ARM_ClearPendingFlag(int ISRIndex) {

#if OS_DEBUG
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return;
  }
#endif
  OS_DI();
  OS_GIC_ClrPending(ISRIndex);
  OS_RestoreI();
}

/*********************************************************************
*
*       OS_ARM_IsPending()
*/
unsigned int OS_ARM_IsPending(int ISRIndex) {
  OS_U32 Result;

#if OS_DEBUG
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return 0;
  }
#endif
  OS_DI();
  Result = OS_GIC_IsPending(ISRIndex);
  OS_RestoreI();
  return Result;
}

/*********************************************************************
*
*       __low_level_init()
*
*       Initialize memory controller, clock generation and PLL
*
*       Has to be modified, if another CPU clock frequency should be
*       used. This function is called during startup and
*       has to return 1 to perform segment initialization.
*
*       Because variables are not initialized before __low_level_init() is called,
*       no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif

OS_INTERWORK int __low_level_init(void) {
  //
  // Initially disable watchdog
  //
  WT_CSR = 0x18u;  // Set to default value (after reset), disabled, interval timer mode
  //
  // Initialize PLL to run specified frequency and speed up system initialization
  //
  _InitPLL();
  //
  // Initialize memory interface for CS0
  //
  BSC_CS0BCR = 0x10000C00u;
  BSC_CS0WCR = 0x00000B40u;
  //
  // Copy code from external Flash into internal RAM if COPY_TO_SRAM is set
  #if COPY_TO_SRAM
    OS_MEMCPY((void*) 0x60100000u, (const void*) 0x0u, 0x900000u);
  #endif
  //
  // Initialize MMU and caches
  //
  //
  // Init MMU and caches. This defines the virtual memory map, which is used during execution.
  // Memory mapping should be complete, meaning 4096 entries.
  // Code below fills in ascending VAddr order
  //
  OS_ARM_MMU_InitTT(_TranslationTable);
  //                                         Mode                      VAddr  PAddr  Size[MB]
#if (COPY_TO_SRAM == 0)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x000, 0x000, 0x040);    // CS0 space, 64MB NOR Flash (16bit BUS on Eval-Board)
#else
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_B,     0x000, 0x601, 0x009);    // Internal SRAM remapped to 0
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x009, 0x009, 0x037);    // CS0 space, 64MB NOR Flash (16bit BUS on Eval-Board)
#endif
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x040, 0x040, 0x040);    // CS1 space, 64MB NOR Flash (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_B,     0x080, 0x080, 0x040);    // CS2 space, 64MB SDRAM (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_B,     0x0C0, 0x0C0, 0x040);    // CS3 space, 64MB SDRAM (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_ILLEGAL, 0x100, 0x100, 0x040);    // CS4 space, User area (not used on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_ILLEGAL, 0x140, 0x140, 0x040);    // CS5 space, User area (not used on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x180, 0x180, 0x040);    // SPI multi I/O BUS space 1, Serial Flash, 64MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x1C0, 0x1C0, 0x040);    // SPI multi I/O BUS space 2, Serial Flash, 64MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_NB,    0x200, 0x600, 0x00A);    // Map the internal RAM from the MIRROR area, DON'T use the physical RAM here.
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x20A, 0x600, 0x00A);    // Map the internal RAM
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_ILLEGAL, 0x214, 0x214, 0x1EB);    // Reserved area
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x3FF, 0x3FF, 0x001);    // L2 Cache
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x400, 0x400, 0x040);    // CS0 mirror space, 64MB NOR Flash (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x440, 0x440, 0x040);    // CS1 mirror space, 64MB NOR Flash (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x480, 0x480, 0x040);    // CS2 mirror space, 64MB SDRAM (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x4C0, 0x4C0, 0x040);    // CS3 mirror space, 64MB SDRAM (16bit BUS on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_ILLEGAL, 0x500, 0x500, 0x040);    // CS4 mirror space, User area (not used on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_ILLEGAL, 0x540, 0x540, 0x040);    // CS5 mirror space, User area (not used on Eval-Board)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x580, 0x580, 0x040);    // SPI multi I/O BUS mirror space 1, Serial Flash, 64MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x5C0, 0x5C0, 0x040);    // SPI multi I/O BUS mirror space 2, Serial Flash, 64MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x600, 0x600, 0x001);    // 1MB internal RAM for RTT
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_C_B,     0x601, 0x601, 0x009);    // 9MB internal RAM mirror space, used as RAM for application
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_NC_NB,   0x60A, 0x60A, 0x9F6);    // SFRs or reserved area, non cacheable, non bufferable
  //
  // Set API functions and base address for L2 Cache
  //
  OS_ARM_AddL2Cache(&OS_L2CACHE_L2C310, (void*)L2CACHE_BASE_ADDR);
  //
  // Now enable MMU, Caches and branch prediction unit
  //
  OS_ARM_MMU_Enable(_TranslationTable);
  OS_ARM_ICACHE_Enable();
  OS_ARM_DCACHE_Enable();
  //
  // Make sure that RTT area is zero-initialized
  //
  memset((void*)0x60020000, 0, 0xE0000);

  return 1;               // Always return 1 to enable segment initilization !
}

/****** End Of File *************************************************/
